/*
    Title:      Wrap-arounds for I/O functions

	Copyright (c) 2000
		Cambridge University Technical Services Limited

	Copying and use of this software is subject to the terms of the
	Poly/ML licence.
*/


#include <errno.h>
#include <sys/types.h> /* Freebsd 2.1 at least requires this before stat.h */
#include <sys/stat.h>


#if defined(HPUX) && !defined(HPUX64)

#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <assert.h>
#include <stdarg.h>

#if defined(HPUX64)
#include <sys/ioctl.h>
#endif

/* The Kernel allows I/O calls to be interrupted, but we don't
   want that, so we'll have to replace them all here.
*/

/******************************************************************************/
/*                                                                            */
/*      LEVEL 2 (File ID) functions                                           */
/*                                                                            */
/******************************************************************************/


/******************************************************************************/
/*                                                                            */
/*      proper_close                                                          */
/*                                                                            */
/******************************************************************************/
int proper_close(int fildes)
{
  int res;
  
  do 
  {
    errno = 0;
    res = close(fildes);
  } while (res == -1 && errno == EINTR);
  
  return res;
}

/******************************************************************************/
/*                                                                            */
/*      proper_open                                                           */
/*                                                                            */
/******************************************************************************/
int proper_open(const char *path, int flag, mode_t mode)
{
  int res;
  
  do 
  {
    errno = 0;
    res = open(path, flag, mode);
  } while (res == -1 && errno == EINTR);
  
  return res;
}

/******************************************************************************/
/*                                                                            */
/*      proper_dup2                                                           */
/*                                                                            */
/******************************************************************************/
int proper_dup2(int fildes, int fildes2)
{
  int res;
  
  do 
  {
    errno = 0;
    res = dup2(fildes, fildes2);
  } while (res == -1 && errno == EINTR);
  
  return res;
}

/******************************************************************************/
/*                                                                            */
/*      proper_ioctl                                                          */
/*                                                                            */
/******************************************************************************/
/* The ioctl call can actualy take parameters of several different types;
   fortunately, we only (currently) use it with one kind of parameter.
   SPF 10/3/97
*/
int proper_ioctl(int fildes, int request, long *arg)
{
  int res;
  
  do 
  {
    errno = 0;
    res = ioctl(fildes, request, arg);
  } while (res == -1 && errno == EINTR);
  
  return res;
}

/******************************************************************************/
/*                                                                            */
/*      proper_fcntl                                                          */
/*                                                                            */
/******************************************************************************/
/* The fcntl call can actually take parameters of several different types;
   fortunately, we only (currently) use it with one kind of parameter.
   SPF 10/3/97
*/
int proper_fcntl(int fildes, int request, int arg)
{
  int res;
  
  do 
  {
    errno = 0;
    res = fcntl(fildes, request, arg);
  } while (res == -1 && errno == EINTR);
  
  return res;
}

/******************************************************************************/
/*                                                                            */
/*      proper_read                                                           */
/*                                                                            */
/******************************************************************************/
int proper_read(int fildes, void *buf, size_t nbyte)
{
  int read_so_far = 0; 

  while (read_so_far < nbyte)
  {
    void *read_from   = (void *)(((char *)buf) + read_so_far);
    size_t read_count = nbyte - read_so_far;
    int res;
    do 
    {
      errno = 0;
      res = read(fildes, read_from, read_count);
    } while (res == -1 && errno == EINTR);
    
    if (res < 0) /* unfixable error */
    {
      return res;
    }
    else if (res == 0) /* end of file reached */
    {
      return read_so_far;
    }
    else read_so_far += res;
  }
  
  return read_so_far;
}

/******************************************************************************/
/*                                                                            */
/*      proper_write                                                          */
/*                                                                            */
/******************************************************************************/
int proper_write(int fildes, const void *buf, size_t nbyte)
{
  int written_so_far = 0; 

  while (written_so_far < nbyte)
  {
    void *write_from   = (void *)(((char *)buf) + written_so_far);
    size_t write_count = nbyte - written_so_far;
    int res;
    do 
    {
      errno = 0;
      res = write(fildes, write_from, write_count);
    } while (res == -1 && errno == EINTR);
    
    if (res < 0) /* unfixable error */
    {
      return res;
    }
    else if (res == 0) /* some resource limit reached */
    {
      return written_so_far;
    }
    else written_so_far += res;
  }
  
  return written_so_far;
}


/******************************************************************************/
/*                                                                            */
/*      LEVEL 3 (Stream) functions                                            */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/*                                                                            */
/*      proper_fopen                                                          */
/*                                                                            */
/******************************************************************************/
FILE *proper_fopen(const char *pathname, const char* type)
{
  FILE *res;
  
  do 
  {
    errno = 0;
    res = fopen(pathname, type);
  } while (res == 0 && errno == EINTR);
  
  return res;
}

/******************************************************************************/
/*                                                                            */
/*      proper_fclose                                                         */
/*                                                                            */
/******************************************************************************/
int proper_fclose(FILE *stream)
{
  int res;
  
  do 
  {
    errno = 0;
    res = fclose(stream);
  } while (res == EOF && errno == EINTR);
  
  return res;
}


/******************************************************************************/
/*                                                                            */
/*      proper_getc                                                           */
/*                                                                            */
/******************************************************************************/
int proper_getc(FILE *stream)
{
  int res;
  do
  {  
    errno = 0;
    res = getc(stream);
  } while (res == EOF && errno == EINTR);

  return res;
}

/* We make ALL the writes unbuffered, so we don't have to call fflush,
   which I think is unreliable (if the underlying write gets interrupted,
   we don't seem to be able to safely restart fflush).
   SPF 19/3/97
*/

/******************************************************************************/
/*                                                                            */
/*      proper_fflush                                                         */
/*                                                                            */
/******************************************************************************/
int proper_fflush(FILE *stream)
{
  int res;
  
  /* Does this retry work? Or has the C library already lost the buffer? */
  do 
  {
    errno = 0;
    res = fflush(stream);
  } while (res == EOF && errno == EINTR);
  
  return res;
}


/******************************************************************************/
/*                                                                            */
/*      proper_putc                                                           */
/*                                                                            */
/******************************************************************************/
/* If the "putc" call is interrupted, we have to restart it "by hand" */
int proper_putc(int c, FILE *stream)
{
  char buffer = c;

  return proper_write(fileno(stream),&buffer,1);
}

/******************************************************************************/
/*                                                                            */
/*      proper_fputs                                                          */
/*                                                                            */
/******************************************************************************/
int proper_fputs(char *s, FILE *stream)
{
  return proper_write(fileno(stream),s,strlen(s));
}

/******************************************************************************/
/*                                                                            */
/*      proper_fwrite_chars                                                   */
/*                                                                            */
/******************************************************************************/
int proper_fwrite_chars(void *s, size_t nchars, FILE *stream)
{
  return proper_write(fileno(stream),s,nchars);
}


/******************************************************************************/
/*                                                                            */
/*      proper_fprintf                                                        */
/*                                                                            */
/******************************************************************************/
int proper_fprintf(FILE *stream, const char *format, ...)
{
  char buffer[BUFSIZ];
  int size;
  
  /* The stdargs macros initialise using the parameter BEFORE the "..." */
  va_list args;
  va_start(args,format);
  size   = vsprintf(buffer, format, args);
  va_end(args);
  
  assert(size < sizeof(buffer));
  
  return proper_write(fileno(stream),buffer,size);
}

/******************************************************************************/
/*                                                                            */
/*      proper_printf                                                         */
/*                                                                            */
/******************************************************************************/
int proper_printf(const char *format, ...)
{
  char buffer[BUFSIZ];
  int size;
  
  /* The stdargs macros initialise using the parameter BEFORE the "..." */
  va_list args;
  va_start(args,format);
  size   = vsprintf(buffer, format, args);
  va_end(args);
 
  assert(size < sizeof(buffer));
  
  return proper_write(fileno(stdout),buffer,size);
}


#endif /* HPUX */


/******************************************************************************/
/*                                                                            */
/*      proper_stat                                                           */
/*                                                                            */
/******************************************************************************/
int proper_stat(char *filename, struct stat *fbp)
{
  int res = stat(filename, fbp);
  while (res != 0 && errno == EINTR) 
  { 
    res = stat(filename, fbp);
    /* There's a bug in SunOS 5.3 (at least) whereby the stat following
       an interrupted stat call can give junk results (ENOENT on a real file).
       I suspect a timing-related problem on automounted file systems.
       So let's try one more time, just to be sure.
       SPF 16/12/96
    */
    if (res != 0)
    {
      res = stat(filename, fbp);
    }
  };
  return res;
}

#ifndef WINDOWS_PC
/* I don't know whether the same applies to lstat but we'll define
   it in the same way just in case. DCJM 16/5/00. */
int proper_lstat(char *filename, struct stat *fbp)
{
  int res = lstat(filename, fbp);
  while (res != 0 && errno == EINTR) 
  { 
    res = lstat(filename, fbp);
    if (res != 0)
    {
      res = lstat(filename, fbp);
    }
  };
  return res;
}
#endif

